@@ -8,79 +8,100 @@ import sqlite3
 
 from mopidy.models import Artist, Album, Track, Ref
 
-_BROWSE_FILTERS = {
-    None: {
-        'album': 'track.album = ?',
-        'artist': 'track.artists = ?',
-        'composer': 'track.composers = ?',
-        'performer': 'track.performers = ?',
-        'genre': 'track.genre = ?',
-        'date': "track.date LIKE ? || '%'"
-    },
-    Ref.ARTIST: {
-        'artist': """
-        EXISTS (SELECT * FROM track WHERE track.artists = artist.uri)
-            OR
-        EXISTS (SELECT * FROM album WHERE album.artists = artist.uri)
-        """,
-        'composer': """
-        EXISTS (SELECT * FROM track WHERE track.composers = artist.uri)
-        """,
-        'performer': """
-        EXISTS (SELECT * FROM track WHERE track.performers = artist.uri)
-        """,
-    },
-    Ref.ALBUM: {
-        'artist': """
-        ? IN (
-            SELECT artists
-             UNION
-            SELECT artists FROM track WHERE album = album.uri
-        )
-        """,
-        'composer': """
-        ? IN (SELECT composers FROM track WHERE album = album.uri)
-        """,
-        'performer': """
-        ? IN (SELECT performers FROM track WHERE album = album.uri)
-        """,
-        'genre': """
-        ? IN (SELECT genre FROM track WHERE album = album.uri)
-        """,
-        'date': """
-        EXISTS (
-            SELECT * FROM track WHERE album = album.uri AND date LIKE ? || '%'
-        )
-        """
-    },
-    Ref.TRACK: {
-        'album': 'album = ?',
-        'artist': """
-        ? IN (
-            SELECT artists
-             UNION
-            SELECT artists FROM album WHERE uri = track.album
-        )
-        """,
-        'composer': 'composers = ?',
-        'performer': 'performers = ?',
-        'genre': 'genre = ?',
-        'date': "date LIKE ? || '%'"
-    }
+_BROWSE_QUERIES = {
+    None: """
+    SELECT CASE WHEN album.uri IS NULL THEN '%s' ELSE '%s' END AS type,
+           coalesce(album.uri, track.uri) AS uri,
+           coalesce(album.name, track.name) AS name
+      FROM track LEFT OUTER JOIN album ON track.album = album.uri
+     WHERE %%s
+     GROUP BY coalesce(album.uri, track.uri)
+     ORDER BY %%s
+    """ % (Ref.TRACK, Ref.ALBUM),
+    Ref.ALBUM: """
+    SELECT '%s' AS type, uri AS uri, name AS name
+      FROM album
+     WHERE %%s
+     ORDER BY %%s
+    """ % Ref.ALBUM,
+    Ref.ARTIST: """
+    SELECT '%s' AS type, uri AS uri, name AS name
+      FROM artist
+     WHERE %%s
+     ORDER BY %%s
+    """ % Ref.ARTIST,
+    Ref.TRACK: """
+    SELECT '%s' AS type, uri AS uri, name AS name
+      FROM track
+     WHERE %%s
+     ORDER BY %%s
+    """ % Ref.TRACK
 }
 
-_BROWSE_SQL = """
-SELECT CASE WHEN album.uri IS NULL THEN 'track' ELSE 'album' END AS type,
-       coalesce(album.uri, track.uri) AS uri,
-       coalesce(album.name, track.name) AS name
-  FROM track LEFT OUTER JOIN album ON track.album = album.uri
+_BROWSE_FILTERS = {
+    None: dict(
+        album='track.album = ?',
+        albumartist='album.artists = ?',
+        artist='track.artists = ?',
+        composer='track.composers = ?',
+        date="track.date LIKE ? || '%'",
+        genre='track.genre = ?',
+        performer='track.performers = ?'
+    ),
+    Ref.ARTIST: dict(
+        role={
+            'albumartist': """EXISTS (
+                SELECT * FROM album WHERE album.artists = artist.uri
+            )""",
+            'artist': """EXISTS (
+                SELECT * FROM track WHERE track.artists = artist.uri
+            )""",
+            'composer': """EXISTS (
+                SELECT * FROM track WHERE track.composers = artist.uri
+            )""",
+            'performer': """EXISTS (
+                SELECT * FROM track WHERE track.performers = artist.uri
+            )"""
+        },
+    ),
+    Ref.ALBUM: dict(
+        albumartist='artists = ?',
+        artist='? IN (SELECT artists FROM track WHERE album = album.uri)',
+        composer='? IN (SELECT composers FROM track WHERE album = album.uri)',
+        date="""EXISTS (
+            SELECT * FROM track WHERE album = album.uri AND date LIKE ? || '%'
+        )""",
+        genre='? IN (SELECT genre FROM track WHERE album = album.uri)',
+        performer='? IN (SELECT performers FROM track WHERE album = album.uri)'
+    ),
+    Ref.TRACK: dict(
+        album='album = ?',
+        albumartist='? IN (SELECT artists FROM album WHERE uri = track.album)',
+        artist='artists = ?',
+        composer='composers = ?',
+        date="date LIKE ? || '%'",
+        genre='genre = ?',
+        performer='performers = ?'
+    )
+}
+
+_SEARCH_SQL = """
+SELECT *
+  FROM tracks
+ WHERE docid IN (SELECT docid FROM %s WHERE %s)
 """
 
-_BROWSE_ALBUMARTIST_SQL = """
-SELECT 'album' AS type, uri AS uri, name AS name
-  FROM album
- WHERE album.artists = ?
-"""
+_SEARCH_FILTERS = {
+    'album': 'album_uri = ?',
+    'albumartist': 'albumartist_uri = ?',
+    'artist': 'artist_uri = ?',
+    'composer': 'composer_uri = ?',
+    'date': "date LIKE ? || '%'",
+    'genre': 'genre = ?',
+    'mtime': "datetime(last_motified, 'unixepoch') LIKE ? || '%'",
+    'performer': 'performer_uri = ?',
+    'uri': 'uri GLOB ?',
+}
 
 _SEARCH_FIELDS = {
     'uri',
@@ -96,22 +117,7 @@ _SEARCH_FIELDS = {
     'comment'
 }
 
-_SEARCH_FILTERS = {
-    'album': 'album_uri = ?',
-    'artist': '? IN (artist_uri, albumartist_uri)',
-    'composer': 'composer_uri = ?',
-    'performer': 'performer_uri = ?',
-    'genre': 'genre = ?',
-    'date': "date LIKE ? || '%'"
-}
-
-_SEARCH_SQL = """
-SELECT *
-  FROM tracks
- WHERE docid IN (SELECT docid FROM %s WHERE %s)
-"""
-
-_USER_VERSION = 3
+schema_version = 4
 
 logger = logging.getLogger(__name__)
 
@@ -128,23 +134,18 @@ class Connection(sqlite3.Connection):
         self.execute('PRAGMA foreign_keys = ON')
         self.row_factory = self.Row
 
-    def executenamed(self, sql, params):
-        sql = sql % (', '.join(params.keys()), ', '.join(['?'] * len(params)))
-        logger.debug('SQLite statement: %s %r', sql, params.values())
-        return self.execute(sql, params.values())
-
 
 def load(c):
-    scripts_dir = os.path.join(os.path.dirname(__file__), b'scripts')
+    sql_dir = os.path.join(os.path.dirname(__file__), b'sql')
     user_version = c.execute('PRAGMA user_version').fetchone()[0]
     if not user_version:
-        logger.info('Creating SQLite database schema v%s', _USER_VERSION)
-        script = os.path.join(scripts_dir, 'create-v%s.sql' % _USER_VERSION)
+        logger.info('Creating SQLite database schema v%s', schema_version)
+        script = os.path.join(sql_dir, 'create-v%s.sql' % schema_version)
         c.executescript(open(script).read())
         user_version = c.execute('PRAGMA user_version').fetchone()[0]
-    while user_version != _USER_VERSION:
+    while user_version != schema_version:
         logger.info('Upgrading SQLite database schema v%s', user_version)
-        script = os.path.join(scripts_dir, 'upgrade-v%s.sql' % user_version)
+        script = os.path.join(sql_dir, 'upgrade-v%s.sql' % user_version)
         c.executescript(open(script).read())
         user_version = c.execute('PRAGMA user_version').fetchone()[0]
     return user_version
@@ -157,7 +158,7 @@ def tracks(c):
 def genres(c):
     return itertools.imap(operator.itemgetter(0), c.execute("""
     SELECT DISTINCT genre
-      FROM tracks
+      FROM track
      WHERE genre IS NOT NULL
      ORDER BY genre
     """))
@@ -166,16 +167,18 @@ def genres(c):
 def dates(c, format='%Y-%m-%d'):
     return itertools.imap(operator.itemgetter(0), c.execute("""
     SELECT DISTINCT strftime(?, date) AS date
-      FROM tracks
+      FROM track
      WHERE date IS NOT NULL
      ORDER BY date
     """, [format]))
 
 
 def images(c):
-    for row in c.execute('SELECT images FROM album WHERE images IS NOT NULL'):
-        for image in row[0].split():
-            yield image
+    return itertools.chain.from_iterable(row[0].split() for row in c.execute("""
+    SELECT DISTINCT images
+      FROM album
+     WHERE images IS NOT NULL
+    """))
 
 
 def lookup(c, uri):
@@ -186,39 +189,39 @@ def lookup(c, uri):
         return None
 
 
+def exists(c, uri):
+    rows = c.execute('SELECT EXISTS(SELECT * FROM track WHERE uri = ?)', [uri])
+    return rows.fetchone()[0]
+
+
 def browse(c, type=None, order=('type', 'name'), **kwargs):
-    if type:
-        sql = "SELECT '%s' AS type, uri, name FROM %s" % (type, type)
-    else:
-        sql = _BROWSE_SQL
-    filters, params = _make_filters(_BROWSE_FILTERS[type], **kwargs)
-    if filters:
-        sql += ' WHERE %s' % ' AND '.join(filters)
-    if not type:
-        sql += ' GROUP BY coalesce(album.uri, track.uri)'
-        # as of sqlite v3.8.2, a UNION seems to be considerably faster
-        # than a WHERE clause spanning multiple tables in a JOIN
-        if 'artist' in kwargs:
-            sql = ' UNION '.join((sql, _BROWSE_ALBUMARTIST_SQL))
-            params.append(kwargs['artist'])
-    if order:
-        sql += ' ORDER BY %s' % ', '.join(order)
-    logger.debug('SQLite query: %s %r', sql, params)
-    return itertools.imap(_ref, c.execute(sql, params))  # imap?
+    filters, params = _filters(_BROWSE_FILTERS[type], **kwargs)
+    sql = _BROWSE_QUERIES[type] % (
+        ' AND '.join(filters) or '1',
+        ', '.join(order)
+    )
+    logger.debug('SQLite browse query %r: %s', params, sql)
+    return [Ref(**row) for row in c.execute(sql, params)]
 
 
-def search_tracks(c, query, limit, offset, exact, **kwargs):
+def search_tracks(c, query, limit, offset, exact, filters=[]):
     if not query:
-        sql, params = ('SELECT * FROM tracks', [])
+        sql, params = ('SELECT * FROM tracks WHERE 1', [])
     elif exact:
-        sql, params = _make_indexed_query(query)
+        sql, params = _indexed_query(query)
     else:
-        sql, params = _make_fulltext_query(query)
-    if kwargs:
-        filters, fparams = _make_filters(_SEARCH_FILTERS, **kwargs)
-        sql = 'SELECT * FROM (%s) WHERE %s' % (sql, ' AND '.join(filters))
-        params.extend(fparams)
-    logger.debug('SQLite query: %s %r', sql, params)
+        sql, params = _fulltext_query(query)
+    clauses = []
+    for kwargs in filters:
+        f, p = _filters(_SEARCH_FILTERS, **kwargs)
+        if f:
+            clauses.append('(%s)' % ' AND '.join(f))
+            params.extend(p)
+        else:
+            logger.debug('Skipped SQLite search filter %r', kwargs)
+    if clauses:
+        sql += ' AND (%s)' % ' OR '.join(clauses)
+    logger.debug('SQLite search query %r: %s', params, sql)
     rows = c.execute(sql + ' LIMIT ? OFFSET ?', params + [limit, offset])
     return map(_track, rows)
 
@@ -229,7 +232,7 @@ def insert_artists(c, artists):
     if len(artists) != 1:
         logger.warn('Ignoring multiple artists: %r', artists)
     artist = next(iter(artists))
-    c.executenamed('INSERT OR REPLACE INTO artist (%s) VALUES (%s)', {
+    _insert(c, 'artist', {
         'uri': artist.uri,
         'name': artist.name,
         'musicbrainz_id': artist.musicbrainz_id
@@ -240,7 +243,7 @@ def insert_artists(c, artists):
 def insert_album(c, album):
     if not album or not album.name:
         return None
-    c.executenamed('INSERT OR REPLACE INTO album (%s) VALUES (%s)', {
+    _insert(c, 'album', {
         'uri': album.uri,
         'name': album.name,
         'artists': insert_artists(c, album.artists),
@@ -254,7 +257,7 @@ def insert_album(c, album):
 
 
 def insert_track(c, track):
-    c.executenamed('INSERT OR REPLACE INTO track (%s) VALUES (%s)', {
+    _insert(c, 'track', {
         'uri': track.uri,
         'name': track.name,
         'album': insert_album(c, track.album),
@@ -311,17 +314,34 @@ def clear(c):
     """)
 
 
-def _make_filters(mapping, role=None, **kwargs):
+def _insert(c, table, params):
+    sql = 'INSERT OR REPLACE INTO %s (%s) VALUES (%s)' % (
+        table,
+        ', '.join(params.keys()),
+        ', '.join(['?'] * len(params))
+    )
+    logger.debug('SQLite insert statement: %s %r', sql, params.values())
+    return c.execute(sql, params.values())
+
+
+def _filters(mapping, role=None, **kwargs):
     filters, params = [], []
-    if role:
-        filters.append(mapping[role])
+    if role and 'role' in mapping:
+        rolemap = mapping['role']
+        if isinstance(role, basestring):
+            filters.append(rolemap[role])
+        else:
+            filters.append(' OR '.join(rolemap[r] for r in role))
     for key, value in kwargs.items():
-        filters.append(mapping[key])
-        params.append(value)
+        if key in mapping:
+            filters.append(mapping[key])
+            params.append(value)
+        else:
+            logger.debug('Skipped SQLite filter expression: %s=%r', key, value)
     return (filters, params)
 
 
-def _make_indexed_query(query):
+def _indexed_query(query):
     terms = []
     params = []
     for field, value in query:
@@ -335,7 +355,7 @@ def _make_indexed_query(query):
     return (_SEARCH_SQL % ('search', ' AND '.join(terms)), params)
 
 
-def _make_fulltext_query(query):
+def _fulltext_query(query):
     terms = []
     params = []
     for field, value in query:
@@ -401,7 +421,3 @@ def _track(row):
             musicbrainz_id=row.performer_musicbrainz_id
         )]
     return Track(**kwargs)
-
-
-def _ref(row):
-    return Ref(type=row.type, uri=row.uri, name=row.name)
